#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>

#include "PCD8544.h"
#include "usart.h"

#define CLK_HALFTIME		0.25

#define PCD8544_CMD_MODE()	(PORTB &= ~(1 << PCD8544_DC))
#define PCD8544_DATA_MODE()	(PORTB |= (1 << PCD8544_DC))

#define PCD8544_SCE_ON() 	(PORTB &= ~(1 << PCD8544_SCE))
#define PCD8544_SCE_OFF() 	(PORTB |= (1 << PCD8544_SCE))

#define PCD8544_RST_ON()	(PORTB &= ~(1 << PCD8544_RST))
#define PCD8544_RST_OFF()	(PORTB |= (1 << PCD8544_RST))

#define PCD8544_SET_VOP		0x80
#define PCD8544_VOP_MAX		0x7F

#define PCD8544_FUNCTION_SET	0x20
#define PCD8544_EXTENDED_INSTR	0x01
#define PCD8544_BASIC_INSTR	0x00

#define PCD8544_DISPLAY_CONTROL	0x08

#define PCD8544_SET_Y_ADDR	0x40
#define PCD8544_SET_X_ADDR	0x80

static uint8_t cur_vop;
static uint8_t cur_addr;

#undef USART

static void spi_init_master()
{
	DDRB |= (1 << PCD8544_SDIN) | (1 << PCD8544_SCLK) | (1 << PB4);
	DDRB &= ~(1 << PB6);

	/* Clock at fOSC / 16 = 1MHz. */
	SPCR = (1 << SPR0) | (1 << MSTR) | (1 << SPE);
}

static void PCD8544_transmit(uint8_t data)
{
	PCD8544_SCE_ON();

	SPDR = data;
	while (!(SPSR & (1 << SPIF)))
		;

#if 0
	int i;
	uint8_t mask;
	char msg[100];

	mask = 1 << 7;
	for (i = 7; i >= 0; i--) {

		sprintf(msg, "mask = 0x%X\n", mask);
		USART1_puts(msg);

		PORTB &= ~(1 << PCD8544_SCLK);
		if ((data & mask) == 0) {
			sprintf(msg, "0 ");
			USART1_puts(msg);
			PORTB &= ~(1 << PCD8544_SDIN);
		} else {
			sprintf(msg, "1 ");
			USART1_puts(msg);
			PORTB |= (1 << PCD8544_SDIN);
		}
		mask >>= 1;
		_delay_us(CLK_HALFTIME);
		PORTB |= (1 << PCD8544_SCLK);
		_delay_us(CLK_HALFTIME);
	}
#endif

	PCD8544_SCE_OFF();
}

void PCD8544_init()
{
	uint16_t i;

	PCD8544_PIN_BANK |= (1 << PCD8544_RST) | (1 << PCD8544_SCE)
		| (1 << PCD8544_DC);
	//PCD8544_PIN_BANK |= | (1 << PCD8544_SDIN) | (1 << PCD8544_SCLK);
	spi_init_master();

	PCD8544_VCC_BANK |= (1 << PCD8544_VCC);
	PCD8544_VCC_PORT |= (1 << PCD8544_VCC);

	/* Apply reset. */
	PCD8544_RST_ON();
	_delay_ms(10);
	PCD8544_SCE_OFF();
	PCD8544_RST_OFF();

	/* Transmit setup commands. */
	PCD8544_CMD_MODE();
	PCD8544_transmit(PCD8544_FUNCTION_SET | PCD8544_EXTENDED_INSTR
			| PCD8544_ADDRESSING_HORIZONTAL);
	cur_addr = PCD8544_ADDRESSING_HORIZONTAL;

	PCD8544_set_vop(PCD8544_DEFAULT_VOP);
	cur_vop = PCD8544_DEFAULT_VOP;

	/* Temperature coefficient 0. */
	PCD8544_transmit(0x04);

	/* Bias mode 1:48. */
	PCD8544_transmit(0x13);

	PCD8544_transmit(PCD8544_FUNCTION_SET | PCD8544_BASIC_INSTR);
	PCD8544_transmit(PCD8544_DISPLAY_CONTROL | PCD8544_VIDEO_MODE_NORMAL);

	/* Clearing the display. */
	PCD8544_DATA_MODE();
	for (i = 0; i < 504; i++)
		PCD8544_transmit(0x00);
}

uint8_t PCD8544_get_vop()
{
	return cur_vop;
}

void PCD8544_cmd(uint8_t cmd)
{
	PCD8544_CMD_MODE();
	PCD8544_transmit(cmd);
}

void PCD8544_clear()
{
	uint8_t prev_addr = cur_addr;
	uint16_t i;
	uint16_t total_chars = 504;

	PCD8544_cmd(PCD8544_FUNCTION_SET | PCD8544_ADDRESSING_HORIZONTAL);

	PCD8544_DATA_MODE();
	for (i = 0; i < total_chars; i++)
		PCD8544_transmit(0x00);

	PCD8544_cmd(PCD8544_FUNCTION_SET | prev_addr);
}

void PCD8544_data(uint8_t data)
{
	PCD8544_DATA_MODE();
	PCD8544_transmit(data);
}

void PCD8544_stream(const uint8_t data[], const uint8_t len)
{
	uint8_t i;

	PCD8544_DATA_MODE();

	for (i = 0; i < len; i++)
		PCD8544_transmit(data[i]);
}

void PCD8544_set_vop(uint8_t vop)
{
	/* Make sure we don't go past the maximum voltage. */
	cur_vop = vop & PCD8544_VOP_MAX;
	PCD8544_cmd(PCD8544_SET_VOP | cur_vop);
}

void PCD8544_addr_mode(uint8_t addr_mode)
{
	PCD8544_cmd(PCD8544_FUNCTION_SET | addr_mode);
	cur_addr = addr_mode;
}

void PCD8544_goto(uint8_t x, uint8_t y)
{
	if (x > PCD8544_MAX_X)
		x = PCD8544_MAX_X;
	if (y > PCD8544_MAX_Y)
		y = PCD8544_MAX_Y;

	PCD8544_CMD_MODE();
	PCD8544_transmit(PCD8544_SET_X_ADDR | x);
	PCD8544_transmit(PCD8544_SET_Y_ADDR | y);
}

#if 0
int main()
{
	PCD8544_init();
	PCD8544_clear();
	//USART1_8N1_init(57600);

	/*
	PCD8544_data(0x00);
	PCD8544_data(0x00);
	PCD8544_data(0x00);

	PCD8544_data(0x1F);
	PCD8544_data(0x10);
	PCD8544_data(0x10);
	PCD8544_data(0x10);
	PCD8544_data(0x00);

	PCD8544_data(0x1F);
	PCD8544_data(0x11);
	PCD8544_data(0x11);
	PCD8544_data(0x1F);
	PCD8544_data(0x00);

	PCD8544_data(0x03);
	PCD8544_data(0x0C);
	PCD8544_data(0x10);
	PCD8544_data(0x0C);
	PCD8544_data(0x03);
	PCD8544_data(0x00);

	PCD8544_data(0x1F);
	PCD8544_data(0x15);
	PCD8544_data(0x15);
	PCD8544_data(0x15);
	PCD8544_data(0x00);

	PCD8544_data(0x00);
	PCD8544_data(0x00);
	PCD8544_data(0x00);

	PCD8544_data(0x1F);
	PCD8544_data(0x05);
	PCD8544_data(0x05);
	PCD8544_data(0x02);
	PCD8544_data(0x00);

	PCD8544_data(0x1F);
	PCD8544_data(0x02);
	PCD8544_data(0x04);
	PCD8544_data(0x02);
	PCD8544_data(0x1F);
	PCD8544_data(0x00);

	PCD8544_data(0x17);
	*/

	_delay_ms(10000);
	while (1) {
		PCD8544_transmit(0xFF);
		//_delay_ms(5000);
		;
	}

	return EXIT_SUCCESS;
}
#endif
