#include <stdio.h>
#include <stdbool.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "adc.h"
#include "usart.h"

#define ADC_START()		(ADCSRA |= (1 << ADSC))

#define TIMER2_START()		(TCCR2B |= (7 << CS20))

#define TIMER2_STOP()		(TCCR2B &= ~(7 << CS20))

/*
 * The ADC pins are only those from the porta A pins.
 */
#define ADC_DDR			DDRA
#define ADC_PORT		PORTA

#define DETECTION_THRESHOLD	850

#define DARK_THRESHOLD		3
#define LIGHT_THRESHOLD		30

volatile bool detect_light;
volatile bool detect_dark;
volatile bool light_detected;

volatile uint8_t dark_count;
volatile uint8_t light_count;

volatile bool *hand_wave_detected;
volatile bool wait_for_variable_reset;

ISR(TIMER2_COMPA_vect)
{
	uint16_t reading;

	reading = ADC;

	/* Waiting for hand passing over the sensor. */
	if (detect_dark) {
		if (reading < DETECTION_THRESHOLD) {
			dark_count++;
			if (dark_count == DARK_THRESHOLD) {
				wait_for_variable_reset = true;
				*hand_wave_detected = true;
				detect_dark = false;
				dark_count = 0;
			}
		} else {
			dark_count = 0;
		}
	/* Waiting for the user to read the variable. */
	} else if (wait_for_variable_reset && *hand_wave_detected == false) {
		wait_for_variable_reset = false;
		detect_light = true;
	}

	if (detect_light) {
		if (reading >= DETECTION_THRESHOLD) {
			light_count++;
			if (light_count == LIGHT_THRESHOLD) {
				detect_light = false;
				detect_dark = true;
				light_count = 0;
			}
		} else {
			light_count = 0;
		}
	}

	/* Start another measurement. */
	ADC_START();
}

void timer2_init()
{
	/* CTC, top at OCR2A. */
	TCCR2A |= (2<< WGM20);

	/* One timer interrupt every 10 miliseconds. */
	OCR2A = 163;

	/* Interrupt on OCR2A compare match. */
	TIMSK2 |= (1 << OCIE2A);

	/* Prescaler at fOSC / 1024; Counter frequency = 16K. */
	TIMER2_START();
}

void adc_init(uint8_t pin, bool *hand_wave)
{
	*hand_wave = false;
	hand_wave_detected = hand_wave;

	/* Voltage reference AVCC. */
	ADMUX |= (1 << REFS0);

	/* Enable the input pin. */
	ADC_DDR &= ~(1 << pin);
	ADMUX |= (pin << MUX0);

	/* Prescaler at 128, slowest but better accuracy. */
	ADCSRA |= (7 << ADPS0);

	/* Enable the ADC. */
	ADCSRA |= (1 << ADEN);

	detect_light = true;
	detect_dark = wait_for_variable_reset = light_detected = false;
	dark_count = light_count = 0;

#ifndef DEBUG
	timer2_init();
#endif

	ADC_START();
}

void adc_pause()
{
	TIMER2_STOP();
}

void adc_resume()
{
	*hand_wave_detected = false;

	detect_light = true;
	detect_dark = wait_for_variable_reset = light_detected = false;
	dark_count = light_count = 0;

	TIMER2_START();
}

#ifdef DEBUG
int main()
{
	uint16_t result;
	bool detected;
	char msg[100];

	sei();

	adc_init(PA7, &detected);
	USART1_8N1_init(115200);

	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);

	while (1) {
		/*
		if (detected) {
			detected = false;
			PORTD ^= (1 << PD7);
			_delay_ms(1000);
		} else {
			PORTD &= ~(1 << PD7);
		}
		*/

		ADC_START();
		while (ADCSRA & (1 << ADSC))
			;
		result = ADC;

		sprintf(msg, "%u\n", result);
		USART1_puts(msg);

		_delay_ms(1000);
	}
}
#endif
