#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include "PCD8544.h"
#include "usart.h"
#include "charmap.h"
#include "music.h"
#include "adc.h"

#define ON(PIN, PIN_REG)	(((PIN_REG) & (1 << (PIN))) == 0)

#define MENU_INPUT_DELAY	5
#define GAME_INPUT_DELAY	6
#define BALL_DELAY		3
#define TOGGLE_DELAY		10

#define WINNING_SCORE		3
#define PLAYER1_WIN		"P1 WINS!"
#define PLAYER2_WIN		"P2 WINS!"
#define PLAYER_WIN_LEN		(sizeof(PLAYER1_WIN)/sizeof(PLAYER1_WIN[0]))

#define PLAYER1_DDR		DDRA
#define PLAYER1_PORT		PORTA
#define PLAYER1_PIN		PINA
#define PLAYER1_UP		PA2
#define PLAYER1_DOWN		PA1

#define PLAYER2_DDR		DDRA
#define PLAYER2_PORT		PORTA
#define PLAYER2_PIN		PINA
#define PLAYER2_UP		PA3
#define PLAYER2_DOWN		PA4

#define MENU_DDR		DDRA
#define MENU_PORT		PORTA
#define MENU_PIN		PINA
#define MENU_BACK		PA0
#define MENU_SELECT		PA5

#define MENU_MODE		0
#define SINGLE_PLAYER_MODE	1
#define MULTI_PLAYER_MODE	2

#define MAX_Y_PIXELS 		((PCD8544_MAX_Y + 1) * 8)

#define PADDLE_LEN		8
#define PADDLE_MAX_Y		(MAX_Y_PIXELS - PADDLE_LEN)

#define TEXT_NEW_GAME 		"NEW GAME"
#define TEXT_NEW_GAME_LEN	(sizeof(TEXT_NEW_GAME)/sizeof(TEXT_NEW_GAME[0]))
#define TEXT_1_PLAYER 		"1 PLAYER"
#define TEXT_1_PLAYER_LEN	(sizeof(TEXT_1_PLAYER)/sizeof(TEXT_1_PLAYER[0]))
#define TEXT_2_PLAYERS		"2 PLAYERS"
#define TEXT_2_PLAYERS_LEN	(sizeof(TEXT_2_PLAYERS)/sizeof(TEXT_2_PLAYERS[0]))
#define TEXT_BACK		"BACK"
#define TEXT_BACK_LEN		(sizeof(TEXT_BACK)/sizeof(TEXT_BACK[0]))
#define TEXT_SELECT		"SELECT"
#define TEXT_SELECT_LEN		(sizeof(TEXT_SELECT)/sizeof(TEXT_SELECT[0]))

enum direction {
	TOP_LEFT = 0,
	TOP_RIGHT,
	BOTTOM_RIGHT,
	BOTTOM_LEFT
};

struct menu {
	bool draw_menu;
	uint8_t game_mode;
	uint8_t pos_x;
	uint8_t arrow_prev_pos_y;
	uint8_t arrow_pos_y;
	uint8_t arrow_pos_x;
	uint8_t delay;
	uint8_t selected_pos_x;
	uint8_t items;
};

struct player {
	uint8_t score;
	uint8_t score_x;
	uint8_t pixel_y;
	uint8_t pixel_x;
	uint8_t delay;
	uint8_t paddle_len;
};

struct ball {
	uint8_t pixel_y;
	uint8_t pixel_x;
	uint8_t delay;
	enum direction dir;
};

/* Symbol for selected single player or multiplayer mode when in the menu. */
uint8_t selected[] = {0x18, 0x3C, 0x7E, 0x3C, 0x18, 0x00};
/* Symbol for current menu item. */
uint8_t select_arrow[] = {0x18, 0x18, 0x5A, 0x3C, 0x18, 0x00};
/* The net. */
uint8_t net[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
/* The ball. */
uint8_t ball_img[] = {0x00, 0x00, 0x18, 0x18, 0x00, 0x00};

uint8_t mode;

struct menu menu;
struct player player1, player2;
struct ball ball;

bool game_reset;
bool toggle_music;
bool music_on;
uint8_t music_toggle_delay;

void init_menu()
{
	MENU_DDR &= ~(1 << MENU_SELECT);
	MENU_DDR &= ~(1 << MENU_BACK);
	MENU_PORT |= (1 << MENU_SELECT) | (1 << MENU_BACK);

	menu.game_mode = MULTI_PLAYER_MODE;
	menu.draw_menu = true;
	menu.pos_x = 16;
	menu.items = 3;
	menu.arrow_pos_x = 8;
	menu.arrow_prev_pos_y = 0;
	menu.delay = MENU_INPUT_DELAY;
	menu.selected_pos_x = 77;
}

static inline void draw_arrow()
{
	PCD8544_goto(menu.arrow_pos_x, menu.arrow_prev_pos_y);
	PCD8544_stream(encode(' '), CHARMAP_CHAR_LEN);

	PCD8544_goto(menu.arrow_pos_x, menu.arrow_pos_y);
	PCD8544_stream(select_arrow, CHARMAP_CHAR_LEN);

	menu.arrow_prev_pos_y = menu.arrow_pos_y;
	menu.delay = MENU_INPUT_DELAY;
}

static inline void draw_selected()
{
	if (menu.game_mode == SINGLE_PLAYER_MODE) {
		PCD8544_goto(menu.selected_pos_x, 2);
		PCD8544_stream(encode(' '), CHARMAP_CHAR_LEN);
		PCD8544_goto(menu.selected_pos_x, 1);
	} else {
		PCD8544_goto(menu.selected_pos_x, 1);
		PCD8544_stream(encode(' '), CHARMAP_CHAR_LEN);
		PCD8544_goto(menu.selected_pos_x, 2);
	}
	PCD8544_stream(selected, CHARMAP_CHAR_LEN);

	menu.delay = MENU_INPUT_DELAY;
}

void draw_menu()
{
	uint8_t i;

	PCD8544_clear();

	menu.arrow_pos_y = 0;

	PCD8544_addr_mode(PCD8544_ADDRESSING_HORIZONTAL);

	PCD8544_goto(menu.pos_x, 0);
	for (i = 0; i < TEXT_NEW_GAME_LEN; i++)
		PCD8544_stream(encode(TEXT_NEW_GAME[i]), CHARMAP_CHAR_LEN);

	PCD8544_goto(menu.pos_x, 1);
	for (i = 0; i < TEXT_1_PLAYER_LEN; i++)
		PCD8544_stream(encode(TEXT_1_PLAYER[i]), CHARMAP_CHAR_LEN);

	PCD8544_goto(menu.pos_x, 2);
	for (i = 0; i < TEXT_2_PLAYERS_LEN; i++)
		PCD8544_stream(encode(TEXT_2_PLAYERS[i]), CHARMAP_CHAR_LEN);

	PCD8544_goto(0, 5);
	for (i = 0; i < TEXT_BACK_LEN; i++)
		PCD8544_stream(encode(TEXT_BACK[i]), CHARMAP_CHAR_LEN);
	PCD8544_goto(49, 5);
	for (i = 0; i < TEXT_SELECT_LEN; i++)
		PCD8544_stream(encode(TEXT_SELECT[i]), CHARMAP_CHAR_LEN);

	draw_arrow();
	draw_selected();
}

void init_players()
{
	PLAYER1_DDR &= ~(1 << PLAYER1_UP);
	PLAYER1_DDR &= ~(1 << PLAYER1_DOWN);
	PLAYER1_PORT |= (1 << PLAYER1_UP) | (1 << PLAYER1_DOWN);

	PLAYER2_DDR &= ~(1 << PLAYER2_UP);
	PLAYER2_DDR &= ~(1 << PLAYER2_DOWN);
	PLAYER2_PORT |= (1 << PLAYER2_UP) | (1 << PLAYER2_DOWN);
}

static inline void reset_players()
{
	player1.score = player2.score = 0;
	player1.delay = player2.delay = GAME_INPUT_DELAY;

	player1.pixel_y = 20;
	player1.score_x = 20;
	player1.pixel_x = 0;
	player1.paddle_len = PADDLE_LEN;

	player2.score_x = 58;
	player2.pixel_x = 82;
	if (mode == SINGLE_PLAYER_MODE) {
		player2.pixel_y = 0;
		player2.paddle_len = 84;
	} else {
		player2.pixel_y = 20;
		player2.paddle_len = PADDLE_LEN;
	}
}

void draw_paddle(uint8_t pixel_x, uint8_t pixel_y)
{
	uint8_t y;
	uint8_t remainder;
	uint8_t first_block, second_block;

	y = pixel_y >> 3;
	remainder = pixel_y & ~0xF8;

	first_block = 0xFF << remainder;
	second_block = 0xFF >> (PADDLE_LEN - remainder);

	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);

	PCD8544_goto(pixel_x, y);
	PCD8544_data(first_block);
	PCD8544_data(second_block);

	PCD8544_goto(pixel_x+1, y);
	PCD8544_data(first_block);
	PCD8544_data(second_block);
}

void draw_full_paddle(uint8_t pixel_x)
{
	uint8_t paddle[6] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);

	PCD8544_goto(pixel_x, 0);
	PCD8544_stream(paddle, 6);
	PCD8544_stream(paddle, 6);
}

void delete_paddle(uint8_t pixel_x, uint8_t pixel_y)
{
	uint8_t y;

	y = pixel_y >> 3;

	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);
	PCD8544_goto(pixel_x, y);
	PCD8544_data(0x00);
	PCD8544_data(0x00);
	PCD8544_goto(pixel_x + 1, y);
	PCD8544_data(0x00);
	PCD8544_data(0x00);
}

static inline void draw_score()
{
	PCD8544_addr_mode(PCD8544_ADDRESSING_HORIZONTAL);

	PCD8544_goto(player1.score_x, 0);
	PCD8544_stream(encode(player1.score+48), CHARMAP_CHAR_LEN);

	PCD8544_goto(player2.score_x, 0);
	PCD8544_stream(encode(player2.score+48), CHARMAP_CHAR_LEN);
}

void draw_net()
{
	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);
	PCD8544_goto(42, 0);
	PCD8544_stream(net, 6);
}

static inline void reset_ball()
{
	ball.pixel_y = 20;
	ball.pixel_x = 42;
	ball.dir = rand() % 4;
	ball.delay = BALL_DELAY;
}

static void draw_ball()
{
	uint8_t outer_y, inner_y, pixel_y;
	uint8_t r;
	uint8_t outer_block1, outer_block2;
	uint8_t inner_block1, inner_block2;

	pixel_y = ball.pixel_y - 1;
	outer_y = pixel_y >> 3;
	r = pixel_y & ~0xF8;

	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);

	outer_block1 = 0x03 << r;
	outer_block2 = 0x03 >> (8-r);
	PCD8544_goto(ball.pixel_x-1, outer_y);
	PCD8544_data(outer_block1);
	PCD8544_data(outer_block2);

	pixel_y = ball.pixel_y - 2;
	inner_y = pixel_y >> 3;
	r = pixel_y & ~0xF8;

	inner_block1 = 0x0F << r;
	inner_block2 = 0x0F >> (8-r);
	PCD8544_goto(ball.pixel_x, inner_y);
	PCD8544_data(inner_block1);
	PCD8544_data(inner_block2);

	PCD8544_goto(ball.pixel_x+1, outer_y);
	PCD8544_data(outer_block1);
	PCD8544_data(outer_block2);
}

static void delete_ball()
{
	uint8_t y, pixel_y;

	pixel_y = ball.pixel_y - 2;
	y = pixel_y >> 3;

	PCD8544_addr_mode(PCD8544_ADDRESSING_VERTICAL);

	PCD8544_goto(ball.pixel_x-1, y);
	PCD8544_data(0x00);
	PCD8544_data(0x00);

	PCD8544_goto(ball.pixel_x, y);
	PCD8544_data(0x00);
	PCD8544_data(0x00);

	PCD8544_goto(ball.pixel_x+1, y);
	PCD8544_data(0x00);
	PCD8544_data(0x00);
}

void update_ball()
{
	ball.delay--;
	if (ball.delay > 0)
		return;

	delete_ball();

	if (ball.dir == TOP_RIGHT) {
		if (ball.pixel_x == player2.pixel_x-1) {
			player1.score++;
			reset_ball();
		} else if (ball.pixel_x == player2.pixel_x-2
				&& ball.pixel_y >= player2.pixel_y
				&& ball.pixel_y < player2.pixel_y+player2.paddle_len) {
			/* Ricochet. */
			ball.dir = TOP_LEFT;
		} else if (ball.pixel_y == 2) {
			ball.dir = BOTTOM_RIGHT;
		} else {
			ball.pixel_x++;
			ball.pixel_y--;
		}
	} else if (ball.dir == BOTTOM_RIGHT) {
		if (ball.pixel_x == player2.pixel_x-1) {
			player1.score++;
			reset_ball();
		} else if (ball.pixel_x == player2.pixel_x-2
				&& ball.pixel_y >= player2.pixel_y
				&& ball.pixel_y < player2.pixel_y+player2.paddle_len) {
			/* Ricochet. */
			ball.dir = BOTTOM_LEFT;
		} else if (ball.pixel_y == MAX_Y_PIXELS-2) {
			ball.dir = TOP_RIGHT;
		} else {
			ball.pixel_x++;
			ball.pixel_y++;
		}
	} else if (ball.dir == BOTTOM_LEFT) {
		if (ball.pixel_x == player1.pixel_x+2) {
			player2.score++;
			reset_ball();
		} else if (ball.pixel_x == player1.pixel_x+3
				&& ball.pixel_y >= player1.pixel_y
				&& ball.pixel_y < player1.pixel_y+player1.paddle_len) {
			/* Ricochet. */
			ball.dir = BOTTOM_RIGHT;
		} else if (ball.pixel_y == MAX_Y_PIXELS-2) {
			ball.dir = TOP_LEFT;
		} else {
			ball.pixel_x--;
			ball.pixel_y++;
		}
	} else if (ball.dir == TOP_LEFT) {
		if (ball.pixel_x == player1.pixel_x+2) {
			player2.score++;
			reset_ball();
		} else if (ball.pixel_x == player1.pixel_x+3
				&& ball.pixel_y >= player1.pixel_y
				&& ball.pixel_y < player1.pixel_y+player1.paddle_len) {
			/* Ricochet. */
			ball.dir = TOP_RIGHT;
		} else if (ball.pixel_y == 2) {
			ball.dir = BOTTOM_LEFT;
		} else {
			ball.pixel_x--;
			ball.pixel_y--;
		}
	}

	draw_ball();

	ball.delay = BALL_DELAY;
}

int main(void)
{
	uint8_t i;

	PCD8544_init();
	adc_init(PA7, &toggle_music);

	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);

	mode = MENU_MODE;
	srand(time(NULL));

	init_menu();
	init_players();
	reset_players();
	reset_ball();

#ifdef DEBUG
	USART1_8N1_init(57600);
#endif

	sei();
	start_music();
	music_on = true;
	music_toggle_delay = TOGGLE_DELAY;

	while (1) {
		if (mode == MENU_MODE) {
			if (menu.draw_menu) {
				pause_music();
				draw_menu();
				menu.draw_menu = false;
			}

			if (ON(PLAYER1_UP, PLAYER1_PIN)
					&& menu.arrow_pos_y > 0) {
				menu.delay--;
				if (menu.delay == 0) {
					menu.arrow_pos_y--;
					draw_arrow();
				}
			} else if (ON(PLAYER1_DOWN, PLAYER1_PIN)
					&& menu.arrow_pos_y < menu.items-1) {
				menu.delay--;
				if (menu.delay == 0) {
					menu.arrow_pos_y++;
					draw_arrow();
				}
			} else if (ON(MENU_SELECT, MENU_PIN)) {
				menu.delay--;
				if (menu.delay == 0) {
					if (menu.arrow_pos_y == 0) {
						/* New game selected. */
						PCD8544_clear();
						if (music_on)
							resume_music();
						adc_resume();
						game_reset = true;

						mode = menu.game_mode;
						menu.delay = MENU_INPUT_DELAY;
					} else {
						if (menu.arrow_pos_y == 1) {
							if (menu.game_mode != SINGLE_PLAYER_MODE)
								game_reset = true;
							menu.game_mode = SINGLE_PLAYER_MODE;
						} else {
							if (menu.game_mode != MULTI_PLAYER_MODE)
								game_reset = true;
							menu.game_mode = MULTI_PLAYER_MODE;
						}
						draw_selected();
					}
				}
			} else if (ON(MENU_BACK, MENU_PIN)) {
				menu.delay--;
				if (menu.delay == 0) {
					PCD8544_clear();
					if (music_on)
						resume_music();
					adc_resume();
					mode = menu.game_mode;
					menu.delay = MENU_INPUT_DELAY;
				}
			}
		} else {
			if (game_reset) {
				game_reset = false;
				reset_players();
				reset_ball();
			}

			if (toggle_music) {
				toggle_music = false;
				if (music_on)
					pause_music();
				else
					resume_music();
				music_on = !music_on;
			}

			update_ball();

			if (ON(PLAYER1_UP, PLAYER1_PIN)
					&& player1.pixel_y > 0) {
				delete_paddle(player1.pixel_x, player1.pixel_y);
				player1.pixel_y--;
			} else if (ON(PLAYER1_DOWN, PLAYER1_PIN)
					&& player1.pixel_y < PADDLE_MAX_Y) {
				delete_paddle(player1.pixel_x, player1.pixel_y);
				player1.pixel_y++;
			}

			if (mode == MULTI_PLAYER_MODE) {
				if (ON(PLAYER2_UP, PLAYER2_PIN)
						&& player2.pixel_y > 0) {
					delete_paddle(player2.pixel_x, player2.pixel_y);
					player2.pixel_y--;
				} else if (ON(PLAYER2_DOWN, PLAYER2_PIN)
						&& player2.pixel_y < PADDLE_MAX_Y) {
					delete_paddle(player2.pixel_x, player2.pixel_y);
					player2.pixel_y++;
				}
			}

			/* Switch back to menu mode. */
			if (ON(MENU_BACK, MENU_PIN)) {
				menu.delay--;
				if (menu.delay == 0) {
					mode = MENU_MODE;
					menu.draw_menu = true;
					menu.delay = MENU_INPUT_DELAY;
					adc_pause();
				}
			}

			draw_net();
			draw_score();
			draw_paddle(player1.pixel_x, player1.pixel_y);

			if (mode == SINGLE_PLAYER_MODE)
				draw_full_paddle(player2.pixel_x);
			else
				draw_paddle(player2.pixel_x, player2.pixel_y);

			if (player1.score == WINNING_SCORE
					|| player2.score == WINNING_SCORE) {

				delete_ball();
				draw_net();

				PCD8544_goto(20, 3);
				PCD8544_addr_mode(PCD8544_ADDRESSING_HORIZONTAL);

				if (player1.score == WINNING_SCORE)
					for (i = 0; i < PLAYER_WIN_LEN; i++)
						PCD8544_stream(encode(PLAYER1_WIN[i]), CHARMAP_CHAR_LEN);
				else
					for (i = 0; i < PLAYER_WIN_LEN; i++)
						PCD8544_stream(encode(PLAYER2_WIN[i]), CHARMAP_CHAR_LEN);
				mode = MENU_MODE;
				menu.delay = MENU_INPUT_DELAY;
				menu.draw_menu = true;
				game_reset = true;
				_delay_ms(2000);
			}
		}

		_delay_ms(30);
	}

	return EXIT_SUCCESS;
}
